import math
import torch
import random


def set_seed(seed):
    torch.manual_seed(seed)
    random.seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def xavier_init(m, seed=None):
    if seed is not None:
        set_seed(seed)

    if isinstance(m, torch.nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight)
        if m.bias is not None:
            torch.nn.init.zeros_(m.bias)
    elif isinstance(m, torch.nn.LayerNorm):
        torch.nn.init.ones_(m.weight)
        torch.nn.init.zeros_(m.bias)
    elif isinstance(m, torch.nn.LSTM):
        for name, param in m.named_parameters():
            if "weight_ih" in name:
                torch.nn.init.xavier_uniform_(param.data)
            elif "weight_hh" in name:
                torch.nn.init.orthogonal_(param.data)
            elif "bias" in name:
                param.data.fill_(0)



import torch
import torch.nn as nn

from typing import Optional

from optrade.torch.models.utils.revin import RevIN
from optrade.torch.models.utils.patcher import Patcher
from optrade.torch.models.utils.pos_enc import PositionalEncoding
from optrade.torch.models.utils.weight_init import xavier_init
from optrade.torch.models.utils.utils import Reshape


# TODO: Reimplement patching
class RecurrentModel(nn.Module):
    def __init__(
        self,
        d_model,
        num_enc_layers,
        pred_len,
        backbone_id,
        bidirectional=False,
        dropout=0.0,
        seq_len=512,
        patching=False,
        patch_dim=16,
        patch_stride=8,
        num_channels=1,
        head_type="linear",
        norm_mode="layer",
        revin=False,
        revout=False,
        revin_affine=False,
        eps_revin=1e-5,
        last_state=True,
        avg_state=False,
        return_head=True,
        channel_independent=False,
        target_channels: Optional[list] = None,
    ) -> None:
        super(RecurrentModel, self).__init__()

        """
        A Recurrent Neural Network (RNN) class that host a variety of different recurrent architectures including LSTM, Mamba, GRU, and the classic RNN.

        Args:
            d_model (int): The number of expected features in the input (required).
            num_enc_layers (int): Number of recurrent layers (required).
            pred_len (int): The number of expected features in the output (required).
            backbone_id (str): The type of recurrent architecture to use (required). Options: "LSTM", "Mamba",
            bidirectional (bool): If True, becomes a bidirectional RNN. Default: False.
            dropout (float): If non-zero, introduces a Dropout layer on the outputs of each RNN layer except the last layer, with dropout probability equal to dropout. Default: 0.
            seq_len (int): The length of the input sequence. Default: 512.
            patching (bool): If True, the input sequence is patched. Default: False.
            patch_dim (int): The dimension of the patch. Default: 16.
            patch_stride (int): The stride of the patch. Default: 8.
            num_channels (int): The number of channels in the input data. Default: 1.
            head_type (str): The type of head to use Options: "linear", "mlp". Default: "linear".
            norm_mode (str): The type of normalization to use. Default: "layer".
            revin (bool): If True, applies RevIN to the input sequence. Default: False.
            revout (bool): If True, applies RevIN to the output sequence. Default: False.
            revin_affine (bool): If True, applies an affine transformation to the RevIN layer. Default: False.
            eps_revin (float): The epsilon value for RevIN. Default: 1e-5.
            last_state (bool): If True, returns the last state of the RNN. Default: True.
            avg_state (bool): If True, returns the average state of the RNN. Default: False.
        """

        # Parameters
        self.backbone_id = backbone_id
        self.num_patches = int((seq_len - patch_dim) / patch_stride) + 2
        self.patch_dim = patch_dim
        self.patch_stride = patch_stride
        self.num_channels = num_channels
        self.eps_revin = eps_revin
        self.revin_affine = revin_affine
        self.revout = revout
        self.target_channels = target_channels
        self.last_state = last_state
        self.avg_state = avg_state
        self.input_size = d_model if patching else num_channels
        self.return_head = return_head

        # RevIN
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        # Patching (only works for FIXED sequence length)
        if patching:
            self._patching = True
            self.patcher = Patcher(patch_dim, patch_stride)
            self.pos_enc = (
                nn.Linear(patch_dim, d_model)
                if avg_state
                else PositionalEncoding(patch_dim, d_model, self.num_patches)
            )
        else:
            self._patching = None

        # Backbone
        if self.backbone_id == "LSTM":
            self.backbone = nn.LSTM(
                input_size=self.input_size,
                hidden_size=d_model,
                num_layers=num_enc_layers,
                batch_first=True,
                dropout=dropout,
                bidirectional=bidirectional,
            )
        elif self.backbone_id == "RNN":
            self.backbone = nn.RNN(
                input_size=self.input_size,
                hidden_size=d_model,
                num_layers=num_enc_layers,
                batch_first=True,
                dropout=dropout,
                bidirectional=bidirectional,
            )
        elif self.backbone_id == "GRU":
            self.backbone = nn.GRU(
                input_size=self.input_size,
                hidden_size=d_model,
                num_layers=num_enc_layers,
                batch_first=True,
                dropout=dropout,
                bidirectional=bidirectional,
            )
        else:
            raise ValueError("Invalid backbone_id. Options: 'LSTM', 'RNN', 'GRU'.")

        # Head
        self.dropout = nn.Dropout(dropout)

        if patching and not avg_state:
            head_dim = self.num_patches * d_model
        elif last_state or avg_state:
            head_dim = d_model

        num_output_channels = (
            len(target_channels) if target_channels is not None else num_channels
        )
        if head_type == "linear":
            self.head = nn.Sequential(
                nn.Linear(head_dim, num_output_channels * pred_len),
                Reshape(-1, num_output_channels, pred_len),
            )
        elif head_type == "mlp":
            self.head = nn.Sequential(
                nn.Linear(head_dim, head_dim // 2),
                nn.GELU(),
                nn.Linear(head_dim // 2, num_output_channels * pred_len),
                Reshape(-1, num_output_channels, pred_len),
            )

        if not (last_state or avg_state):
            self.head = nn.Sequential(
                Reshape(-1, seq_len * d_model),
                self.head,
            )
        self.flatten = nn.Flatten(start_dim=-2)

        # Final Normalization Layer
        norm_dim = d_model
        self.norm = nn.LayerNorm(norm_dim) if norm_mode == "layer" else nn.Identity()

        # Weight initialization
        self.apply(xavier_init)

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def compute_backbone(self, x):
        if self.backbone_id in {"RNN", "GRU"}:
            out, hn = self.backbone(x)
            last_hn = hn[-1]
        elif self.backbone_id == "LSTM":
            out, (hn, _) = self.backbone(x)
            last_hn = hn[-1]
        else:
            raise ValueError("Invalid backbone_id. Options: 'LSTM', 'RNN', 'GRU',.")

        return out, last_hn

    def forward(self, x):
        """
        Forward pass of the model.

        Args:
            x (torch.Tensor): The input data. Shape: (batch_size, num_channels, seq_len) = (B, M, L).
        """

        # Ensure input is correct
        if len(x.shape) == 2:
            x = (
                x.unsqueeze(-2) if self._patching else x.unsqueeze(-1)
            )  #: (B, L) -> (B, L, 1)

        # RevIN
        if self._revin:
            x = self.revin(
                x, mode="norm"
            )  # Patched version:(B, M, L). Non-patched version: (B, L, 1)

        x = x.transpose(1, 2)

        # # Patching
        # if self._patching:
        #     x = self.patcher(x) # (B, M, N, P)
        #     x = self.pos_enc(x) # (B, M, N, D)
        #     B, M, N, D = x.shape
        #     x = x.view(B*M, N, D) # (B*M, N, D)

        # Backbone forward pass
        out, last_hn = self.compute_backbone(x)

        # Normalization
        if self.last_state:
            x = self.norm(last_hn)  # Select last hidden state: (B, D)
        elif self.avg_state:
            x = self.norm(
                torch.mean(out, dim=1)
            )  # Average over sequence length. Patched version: (B*M, D). Non-patched version: (B, D).
        else:
            x = self.norm(
                out
            )  # Patched version: (B*M, N, D). Non-patched version: (B, L, D)

        # # Reshape for patching
        # if self._patching:
        #     x = x.view(B, M, -1) # avg state: (B, M, D). Non-avg state: (B, M, N*D)

        # Head
        if self.return_head:
            # x = x.transpose(0,1)
            print(f"Shape before head: {x.shape}")
            x = self.head(self.dropout(x))  # (B, pred_len)

        print(f"x after head: {x.shape}")
        # RevOUT
        if self.revout:
            x = self.revin(x, mode="denorm")

        return x


if __name__ == "__main__":
    # <---Non-patched version (classification)--->
    # Define model parameters
    batch_size = 32
    num_channels = 7
    seq_len = 512
    pred_len = 96
    d_model = 64
    num_enc_layers = 5

    # Device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(batch_size, num_channels, seq_len).to(device)  # (B, M, L)

    model = RecurrentModel(
        d_model=d_model,
        num_enc_layers=num_enc_layers,
        pred_len=pred_len,
        backbone_id="GRU",
        bidirectional=False,
        dropout=0.1,
        seq_len=seq_len,
        patching=False,
        # patch_dim=16,
        # patch_stride=8,
        num_channels=num_channels,
        head_type="linear",
        norm_mode="layer",
        revin=True,
        revout=True,
        revin_affine=True,
        last_state=False,
        avg_state=True,
        return_head=True,
        target_channels=[0, 3, 5],
    ).to(device)

    # Pass the data through the model
    output = model(x)
    output = output.to(device)

    print(f"Input shape: {x.shape}")
    print(f"Output shape: {output.shape}")

    # #<--Patched Version (forecasting)--->
    # # Define model parameters
    # patch_dim = 64
    # patch_stride = 16
    # batch_size = 1
    # d_model = 128
    # num_enc_layers = 5
    # pred_len = 1
    # seq_len = 16031
    # num_channels = 1

    # # Device
    # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # print(f"Using device: {device}")

    # # Create an instance of the LSTM model
    # model = RecurrentModel(
    #     d_model=d_model,
    #     backbone_id="Mamba",
    #     num_enc_layers=num_enc_layers,
    #     pred_len=pred_len,
    #     seq_len=seq_len,
    #     num_channels=num_channels,
    #     revin=True,
    #     revin_affine=True,
    #     revout=True,
    #     head_type="linear",
    #     patching=True,
    #     last_state=False,
    #     avg_state=True,
    # ).to(device)

    # # Create sample input data
    # x = torch.randn(batch_size, num_channels, seq_len).to(device)  # (B, M, L)
    # print(f"Input shape: {x.shape}")

    # # Pass the data through the model
    # output = model(x)
    # output = output.to(device)

    # print(f"Output shape: {output.shape}")



import os
import json
import argparse
import sys
from datetime import datetime
from optrade.exp.exp import Experiment
from optrade.config.config import load_config
import torch
from dotenv import load_dotenv
from rich.console import Console
from rich.pretty import pprint
import warnings
from pydantic import BaseModel
from typing import Dict, Any

warnings.filterwarnings(
    "ignore", message="h5py not installed, hdf5 features will not be supported."
)


def update_global_config(
    ablation_config: Dict[str, Any], global_config: BaseModel, ablation_id: int
) -> BaseModel:
    """
    Updates the global config for ablation studies and hyperparameter tuning. For example,
    if the ablation_config is {'global_config.sl.lr': 0.01}, then the global_config.sl.lr will be
    updated to 0.01.
    """

    for key, value in ablation_config.items():
        parts = key.split(".")
        if len(parts) == 2:
            sub_model, param = parts
            if hasattr(global_config, sub_model):
                sub_config = getattr(global_config, sub_model)
                if hasattr(sub_config, param):
                    setattr(sub_config, param, value)
                else:
                    print(f"Warning: {sub_model} does not have attribute {param}")
            else:
                print(f"Warning: global_config does not have attribute {sub_model}")
        elif len(parts) == 3:
            model, sub_model, param = parts
            if model == "global" and hasattr(global_config, sub_model):
                sub_config = getattr(global_config, sub_model)
                if hasattr(sub_config, param):
                    setattr(sub_config, param, value)
                else:
                    print(f"Warning: {sub_model} does not have attribute {param}")
            else:
                print(f"Warning: Invalid key format or 'global' not specified: {key}")
        else:
            print(f"Warning: Invalid key format: {key}")

    global_config.exp.ablation_id = ablation_id

    return global_config


def main(job_name="test", ablation=None, ablation_id=1):

    # Rich console
    load_dotenv()
    console = Console()

    # Load experimental configuration
    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    args_path = os.path.join(base_dir, "optrade", "jobs", job_name, "args.yaml")
    args = load_config(args_path)
    if ablation is not None:
        args = update_global_config(ablation, args, ablation_id)
    console.print("Pydantic Configuration Loaded Successfully:", style="bold green")

    # Run experiment on multiple seeds (optional)
    seed_list = args.exp.seed_list

    for i in range(len(seed_list)):
        args.exp.seed = seed_list[i]
        args.exp.time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        # Print args
        pprint(args, expand_all=True)

        # Initialize experiment
        exp = Experiment(args)

        console.log("Using single device")
        exp.run()


if __name__ == "__main__":
    # Non-hyperparameter tuning
    warnings.filterwarnings(
        "ignore", message="h5py not installed, hdf5 features will not be supported."
    )
    parser = argparse.ArgumentParser(description="Run experiment")
    parser.add_argument("job_name", type=str, default="test", help="Name of the job")
    parser.add_argument("--venv_test", action="store_true", help="Flag for venv test")
    parser.add_argument(
        "--ablation", type=str, default=None, help="Ablation study configuration"
    )
    parser.add_argument(
        "--ablation_id", type=int, default=1, help="Ablation study configuration"
    )

    args = parser.parse_args()

    if args.venv_test:
        print("Running in venv test mode")
        # Add any venv test specific code here
    else:
        if args.ablation is not None:
            try:
                ablation = json.loads(args.ablation)
                main(
                    job_name=args.job_name,
                    ablation=ablation,
                    ablation_id=args.ablation_id,
                )
            except json.JSONDecodeError:
                print("Error: Invalid JSON string for ablation configuration")
                sys.exit(1)
        else:
            main(args.job_name)
